/* zklMsgHash.c : Message/Digest Hashes
 * Christophe Devine's SHA-256 & SHA-1 implementation (GPL2)
 * Antoon Bosselaers' RIPEMD-160 implementation (GPL)
 * Chris Venter's libb64 (base64 codec)
 *
 * Libraries you will need: None
 * 
 * Copyright (c) 2014 Craig Durland craigd@zenkinetic.com
 * License: The zlib/libpng License,
 * 	http://www.opensource.org/licenses/zlib-license.php
 */

#define _CRT_SECURE_NO_WARNINGS		// VC++ 2008

#include "zklObject.h"
#include "zklData.h"
//#include "zklMethod.h"
#include "zklString.h"
#include "zklUtil.h"
#include "zklImports.h"

#include <string.h>
#include <stdio.h>
#include "sha1.h"
#include "sha256.h"
#include "ripemd160.h"
#include "cdecode.h"
#include "cencode.h"
#include "base32.h"
#include "MD5/global.h"
#include "MD5/md5.h"

static unsigned char *_getMsg(Instance *text,size_t *sz, pVM vm)
{
   char     *message;

   switch (TYPEO(text))
   {
      case DataType:
	 message = (char *)dataText(text,sz);
	 break;
      case StringType:
	 message = stringText(text); *sz = strlen(message);
	 break;
      default:
      {
	 char	buf[100];
	 sprintf(buf,"MsgHash: Can't handle %s",iname(text));
	 vmThrow(vm,E_TYPE_ERROR,buf);
      }
   }
   return (unsigned char *)message;
}

static unsigned char *getMsg(pArglist arglist,size_t *sz, pVM vm)
   { return _getMsg(arglistGet(arglist,0,"MsgHash",vm),sz,vm); }

static Instance *formatHash(unsigned char *buf,int sz,Instance *data,pVM vm)
{
   if(data && TYPEO(data)==DataType && data!=VData)
   {
      dataWillFill(data);
      dataInsertBytes(data,0,(Byte *)buf,sz,1,vm);
      return data;
   }

   {
      Byte hash[65];
      int i,n;
      for (i = n = 0; i < sz; i++, n += 2)
	 sprintf((char *)&hash[n],"%02x", buf[i]);
      return stringCreate((char *)hash,I_OWNED,vm);
   }
}

    // SHA256(text,N=1 [,Data()])
    // N<1 --> garbage
    // To hash the hash: (n).reduce(SHA256.fp1(1,False),text)
static Instance *SHA256_calc(Instance *self,pArglist arglist,pVM vm)
{
   unsigned char *message;

   int	     N=1;
   int64_t   n64;
   size_t    len;
   sha256_context ctx;
   unsigned char  sha256sum[32];

   if (arglistTryToGetInt(arglist,1,&n64,0,vm)) N = (int)n64;

   message = getMsg(arglist,&len,vm);
   sha256_starts(&ctx);
   do sha256_update(&ctx,(uint8 *)message,len); while(0<--N);
   sha256_finish(&ctx,sha256sum);

   return formatHash(sha256sum,32,arglistTryToGetBT(arglist,2,DataType,0,vm),vm);
}

    // MD5.calc(text)
    // MD5.calc(text,N==1,asText=True)
    // This has been duplicated from the VM
static Instance *MD5_calc(Instance *self,pArglist arglist,pVM vm)
{
   MD5_CTX	context;
   Byte		digest[16];
   size_t	len;
   unsigned char *message = getMsg(arglist,&len,vm);

   MD5Init(&context);
   MD5Update(&context,message,len);
   MD5Final(digest,&context);

//   return formatHash(digest,16,arglistTryToGetBool(arglist,2,1,0,vm),vm);
   return formatHash(digest,16,arglistTryToGetBT(arglist,2,DataType,0,vm),vm);
}

////////////////////////////////////// Sink /////////////////////////////
// Nice for hashing video files so you don't have to load the entire file
// var m=Import("zklMsgHash"), s=m.sink("MD5"), d=Data(0d50_000);
// File(biggie).pump(d,s); s.close();  // chunks the file through d
// Which, on my machine, saves a whole 1.3 seconds on a 1.8GB file

#define SHA256	0
#define MD5	1
typedef struct { int hashType; unsigned char ctx[0]; }Roller;
#define RCTX(rlr) ( (((Roller *)rlr)->ctx) )
static int rollingCalc(void *rlr, Instance *i, pVM vm)
{
   size_t         sz;
   unsigned char *msg = _getMsg(i,&sz,vm);
   void          *ctx = RCTX(rlr);
   switch(((Roller *)rlr)->hashType)
   {
      case SHA256: sha256_update((sha256_context *)ctx,(uint8 *)msg,sz); break;
      case MD5:    MD5Update((MD5_CTX *)ctx,msg,sz);		         break;
      default: return 0;
   }
   return 1;
}

    // called only once, when the sink is closed
    // or after pump 
static Instance *rollingClosed(void *rlr, pVM vm) // hard close
{
   unsigned char buffer[100];
   void		*ctx = RCTX(rlr);
   int		 sz;
   switch(((Roller *)rlr)->hashType)
   {
      case SHA256: sha256_finish((sha256_context *)ctx,buffer); sz = 32; break;
      case MD5:    MD5Final(buffer,(MD5_CTX *)ctx); 		sz = 16; break;
      default: return Void;
   }

   return formatHash(buffer,sz,0,vm);
}

    // MsgHash.toSink("MD5" | "SHA256")
    // Can't use object->toSInk because I don't have an object
    //   Sink(MsgHash,"SHA256")
    // ??? toSInk(MsgHash.SHA256)
static Instance *MsgHash_toSink(Instance *self,pArglist arglist,pVM vm)
{
#if 1
   char     *nm = arglistGetOnlyString(arglist,0,"MsgHash.sink",vm);
#else
   Instance *pm = arglistGetBT(arglist,0,MethodType,"MsgHash.sink",vm);
   pMethod   m  = methodMethod(pm);
#endif
   CSink     cs = { 0, rollingCalc,rollingClosed };
   Roller   *rlr;
   int	     hashType;
   Instance *sink;
   ZAgg     *aggie;

   if(0==strcmp("SHA256",nm)){
//   if(m==SHA256_calc){
      sink = sinkCreate(sizeof(Roller) + sizeof(sha256_context),&aggie,vm);
      sha256_starts((sha256_context *)RCTX(aggie->sandbox));
      hashType = SHA256;
   }
   else if(0==strcmp("MD5",nm)){
//   else if(m==MD5_calc){
      sink = sinkCreate(sizeof(Roller) + sizeof(MD5_CTX),&aggie,vm);
      MD5Init((MD5_CTX *)RCTX(aggie->sandbox));
      hashType = MD5;
   }
//   else return Void;	// !!!ValueError
   else vmThrow(vm,"ValueError","MsgHash.toSink: unknown sink name");

   aggie->how	    = ZA_C;
   rlr = cs.X       = (Roller *)aggie->sandbox;  // in Sink
   rlr->hashType    = hashType;
   aggie->dst.csink = cs;	// struct copy
   return sink;
}

    // SHA1(text,N=1 [,Data()])
    // N<1 --> garbage
    // To hash the hash: (n).reduce(SHA1.fp1(1,False),text)
static Instance *SHA1_calc(Instance *self,pArglist arglist,pVM vm)
{
   unsigned char *message;

   int	     N=1;
   int64_t   n64;
   size_t    len;
   sha1_context  ctx;
   unsigned char sha1sum[20];

   if (arglistTryToGetInt(arglist,1,&n64,0,vm)) N = (int)n64;

   message = getMsg(arglist,&len,vm);
   sha1_starts(&ctx);
   do sha1_update(&ctx,(uint8 *)message,len); while(0<--N);
   sha1_finish(&ctx,sha1sum);

   return formatHash(sha1sum,20,arglistTryToGetBT(arglist,2,DataType,0,vm),vm);
}

    // RIPEMD160(text,N=1 [,Data()])
static Instance *ripemd160_calc(Instance *self,pArglist arglist,pVM vm)
{
   unsigned char *message,hash[30];
   size_t    sz;
   struct RIPEMD160Context ctx;

   message = getMsg(arglist,&sz,vm);
   RIPEMD160_init(&ctx);
   RIPEMD160_add_data(&ctx,message,sz);
   RIPEMD160_finish(&ctx,hash);

   return formatHash(hash,20,arglistTryToGetBT(arglist,2,DataType,0,vm),vm);
}

    // base64decode(message): result is 3/4 size of message
static Instance *base64_decode(Instance *self,pArglist arglist,pVM vm)
{
   size_t sz;
   unsigned char *message = getMsg(arglist,&sz,vm);
   base64_decodestate _state;
   Instance *r=dataCreate(sz*3/4+20,I_OWNED,vm);
   void     *ptr=dataWillFill(r);

   base64_init_decodestate(&_state);
   sz=base64_decode_block((const char *)message,sz, ptr, &_state);
   dataSetSize(r,sz);	    // shrink to fit
   return r;
}

    // base64encode(message) --> Data(plain text)
    // result is bigger than message: 4/3  + padding + '\n' every 72 chars
static Instance *base64_encode(Instance *self,pArglist arglist,pVM vm)
{
   size_t sz,z;
   unsigned char *message = getMsg(arglist,&sz,vm);
   base64_encodestate _state;
   Instance *r;
   void     *ptr;

   z = 4*(sz+2)/3; z += z/72 + 20;
   r = dataCreate(z,I_OWNED,vm); ptr = dataWillFill(r);

   base64_init_encodestate(&_state);
   sz=base64_encode_block((const char *)message,sz,ptr,&_state);
   sz+=base64_encode_blockend((char *)ptr+sz, &_state);
   dataSetSize(r,sz);	    // shrink to fit
   return r;
}

    // base32decode(message): 
static Instance *base32decode(Instance *self,pArglist arglist,pVM vm)
{
   int    n;
   size_t sz, bufSz;
   unsigned char *message = getMsg(arglist,&sz,vm);
   Instance *r;
   void     *ptr;

   bufSz = sz*5/8 + 20;
   r     = dataCreate(bufSz,I_OWNED,vm);
   ptr   = dataWillFill(r);

   n = base32_decode(message,ptr,bufSz);
   if(n<0) vmThrow(vm,E_ASSERTION_ERROR,"base32_decode doesn't like message");
   dataSetSize(r,n);	    // shrink to fit
   return r;
}

    // base32encode(message,pad=True) --> Data(plain text)
    // Result is bigger than message: 160% (8/5)
static Instance *base32encode(Instance *self,pArglist arglist,pVM vm)
{
   int    n;
   size_t sz,z;
   unsigned char *message = getMsg(arglist,&sz,vm);
   Instance *r;
   void     *ptr;

   z = 8*(sz+4)/5; z += z/72 + 20;
   r = dataCreate(z,I_OWNED,vm); ptr = dataWillFill(r);

   n=base32_encode(message,sz,ptr,z,arglistTryToGetBool(arglist,1,1,0,vm));
   if(n<0) vmThrow(vm,E_ASSERTION_ERROR,"base32_encode: message too long");
   dataSetSize(r,n);	    // shrink to fit
   return r;
}

static const MethodTable MsgHashMethods[] =
{
   "SHA256",		(pMethod)SHA256_calc,
   "SHA1",		(pMethod)SHA1_calc,
   "RIPEMD160",		(pMethod)ripemd160_calc,
   "base64decode",	(pMethod)base64_decode,
   "base64encode",	(pMethod)base64_encode,
   "base32encode",	(pMethod)base32encode,
   "base32decode",	(pMethod)base32decode,
   "MD5",		(pMethod)MD5_calc,
   "toBool",		Bool_soTrue,
   "toSink",		MsgHash_toSink,
   0,			0
};

/* ******************************************************************** */
/* **************************** Properties **************************** */
/* ******************************************************************** */

static Instance *extra = 0;  // static immortal class, don't need to mark

    // .extra -->class
static Instance *extraClass(Instance *self, pVM vm){ return extra; }

    // .version -->string
static Instance *version(Instance *self, pVM vm){ return emptyString; }

    // .authors
static Instance *authors(Instance *self, pVM vm)
{
   return createTupleOfStrings(vm,"Christophe Devine (SHA-256)", 
      "Xavier Leroy (RIPEMD-160)", "Chris Venter (libb64)",
      "Markus Gutschke (base32)",
      (char *)0);
}

    // .libraries
static Instance *libraries(Instance *self, pVM vm){ return emptyString; }

static PropertyTable properties[] =
{
   "extra",		(pProperty)extraClass,
   "version",		(pProperty)version,
   "authors",		(pProperty)authors,
   "libraries",		(pProperty)libraries,
   0,			0
};


static ZKL_Object MsgHashObject;
static Instance	  msgHash;

DllExport void *construct(void *vm)
{
   extern char wad[];  // wad.c

   if (!vm) return (void *)ZKLX_PROTOCOL;	// handshake
   	// If this is reloaded, nothing happens except 
	// construct() is called again so don't reinitialize
   if (!MsgHashObject.name)	// static items are zero
   {
      constructObject(&MsgHashObject,NativeType, MsgHashMethods,properties,0,vm);
      MsgHashObject.name        = "MsgHash";
      MsgHashObject.vaultPath   = "Utils";
      MsgHashObject.isize       = sizeof(Instance);

      instanceInit(&msgHash,&MsgHashObject,I_UNTOUCHABLE);

      extra = loadWad((Byte *)wad,0,NoArglist,vm);
   }
   return &msgHash;
}
